/* Copyright (c) 2006 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmladdin.cpp - Workbench extension: editor customizations for HTML/XML
Function
  This is a TADS Workbench add-in that provides customized indenting and
  styling for HTML and XML source files opened in Workbench's integrated
  Scintilla editor.

  See itadsworkbench.h for details on how to create Workbench extensions.
Notes
  
Modified
  10/05/06 MJRoberts  - Creation
*/

#include <string.h>
#include <stdlib.h>
#include <windows.h>

#include <scintilla.h>

#include "wb_addin.h"
#include "htmladdin.h"


/* ------------------------------------------------------------------------ */
/*
 *   Module management 
 */

/* our DLL's instance handle - we need this to load our resources */
HINSTANCE G_instance = 0;

/* main DLL entrypoint */
BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, VOID *reserved)
{
    /* remember our instance handle */
    if (reason == DLL_PROCESS_ATTACH)
        G_instance = hinst;

    /* success */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   XML editor mode 
 */
class XmlEditorMode: public CTadsWorkbenchEditorMode
{
public:
    XmlEditorMode(ITadsWorkbench *itwb)
        : CTadsWorkbenchEditorMode(itwb)
    {
    }

    /*
     *   ITadsWorkbenchEditorMode methods 
     */

    /* get my mode name */
    BSTR STDMETHODCALLTYPE GetModeName()
    {
        return SysAllocString(L"XML");
    }

    /* get my filter string */
    BSTR STDMETHODCALLTYPE GetFileDialogFilter(BOOL for_save, UINT *priority)
    {
        wchar_t str[] = L"XML Files (*.xml)\0*.xml\0";

        *priority = 1001;
        return SysAllocStringLen(str, sizeof(str)/sizeof(str[0]) - 1);
    }

    /* check to see if we want to provide editor customizations for a file */
    UINT STDMETHODCALLTYPE GetFileAffinity(const OLECHAR *filename)
    {
        /* the XML mode recognizes the .xml extension */
        size_t len = wcslen(filename);
        if (len > 4 && wcsicmp(filename + len - 4, L".xml") == 0)
        {
            /* we matched on extension */
            return 100;
        }
        
        /* not one of ours */
        return 0;
    }

    /* get file type information */
    BOOL STDMETHODCALLTYPE GetFileTypeInfo(UINT index, wb_ftype_info *info)
    {
        switch (index)
        {
        case 0:
            /* XML files */
            info->icon = (HICON)LoadImage(
                G_instance, MAKEINTRESOURCE(ICON_XML_FILE),
                IMAGE_ICON, LR_DEFAULTSIZE, LR_DEFAULTSIZE, LR_DEFAULTCOLOR);
            info->need_destroy_icon = FALSE;
            wcscpy(info->name, L"XML File");
            wcscpy(info->desc, L"A file containing XML (eXtensible "
                   L"Markup Language) code");
            wcscpy(info->defext, L"xml");
            return TRUE;

        default:
            /* no more modes */
            return FALSE;
        }
    }

protected:
    /* create our extension */
    CTadsWorkbenchEditor *create_ext(HWND hwnd);

    /* get the number of styles we use */
    virtual const style_info *get_style_array() { return styles_; }
    virtual UINT get_style_count() { return style_cnt_; }

protected:
    /* default colors for the styles */
    static const style_info styles_[];
    static size_t style_cnt_;
};

/* our style list - XML and HTML modes share the style numbering system */
const style_info XmlEditorMode::styles_[] =
{
    { 0, L"Plain Text", L"Sample Text",
    RGB(0x00,0x00,0xFF), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 1, L"HTML Tags", L"<TABLE>",
    RGB(0x80,0x00,0xFF), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 2, L"HTML Unknown Tags", L"<UNDEFINED>",
    RGB(0xFF,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 3, L"HTML Attributes", L"href=",
    RGB(0x00,0x80,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 4, L"HTML Unknown Attributes", L"undefined=",
    RGB(0xFF,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 5, L"HTML Numbers", L"12345",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 6, L"HTML Double quoted strings", L"\"Sample Text\"",
    RGB(0xc0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 7, L"HTML Single quoted strings", L"'Sample Text'",
    RGB(0xc0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 8, L"HTML Other inside tag", L"Sample Text",
    RGB(0x80,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 9, L"HTML Comment", L"<!-- Comment -->",
    RGB(0x80,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 10, L"HTML Entities", L"&&amp;",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 11, L"XML style tag end", L"<LINK/>",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 12, L"XML identifier start", L"<?XML",
    RGB(0x00,0x00,0xFF), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 13, L"XML identifier end", L"?>",
    RGB(0x00,0x00,0xFF), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 14, L"SCRIPT", L"Sample Text",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 15, L"ASP1", L"<% Sample %>",
    RGB(0xFF,0xFF,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 16, L"ASP2", L"<% Sample %>",
    RGB(0xFF,0xD0,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 17, L"HTML CDATA", L"Sample Text",
    RGB(0xFF,0xD0,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 18, L"PHP", L"Sample Text",
    RGB(0x00,0x00,0xFF), RGB(0xFF,0xFF,0xD0), 0, 0 },

    { 19, L"PHP Unquoted values", L"Sample Text",
    RGB(0xFF,0x00,0xFF), RGB(0xFF,0xEF,0xFF), 0, 0 },

    { 20, L"JSP Comment", L"<%-- Comment --%>",
    RGB(0x00,0x00,0x00), RGB(0xFF,0xFF,0xD0), 0, 0 },

    { 21, L"SGML tags", L"<!Sample>",
    RGB(0x00,0x00,0x80), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 22, L"SGML command", L"Sample Text",
    RGB(0x00,0x00,0x80), RGB(0xEF,0xEF,0xFF), WBSTYLE_BOLD, 0 },

    { 23, L"SGML 1st param", L"Sample Text",
    RGB(0x00,0x64,0x00), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 24, L"SGML double string", L"\"Sample Text\"",
    RGB(0x80,0x00,0x00), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 25, L"SGML single string", L"'Sample Text'",
    RGB(0x99,0x33,0x00), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 26, L"SGML error", L"Sample Text",
    RGB(0x80,0x00,0x00), RGB(0xFF,0x66,0x66), 0, 0 },

    { 27, L"SGML special", L"#1234",
    RGB(0x33,0x66,0xFF), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 28, L"SGML entity", L"Sample Text",
    RGB(0x33,0x33,0x33), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 29, L"SGML comment", L"Sample Text",
    RGB(0x80,0x80,0x00), RGB(0xEF,0xEF,0xFF), 0, 0 },

    { 31, L"SGML block", L"Sample Text",
    RGB(0x00,0x00,0x80), RGB(0xC0,0xC0,0xE0), 0, 0 },

    { 34, L"Brace Highlight", L"< >",
    RGB(0x00,0x00,0x00), RGB(0x7F,0xFF,0xD4), 0, 0 },

    { 35, L"Matched Operators", L"",
    RGB(0xFF,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 40, L"JavaScript Start", L"Sample Text",
    RGB(0x80,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 41, L"JavaScript Default", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 42, L"JavaScript Comment", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 43, L"JavaScript Line Comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 44, L"JavaScript Doc comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 45, L"JavaScript Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 46, L"JavaScript Word", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 47, L"JavaScript Keyword", L"Sample Text",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 48, L"JavaScript Double quoted string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 49, L"JavaScript Single quoted string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 50, L"JavaScript Symbols", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 51, L"JavaScript EOL", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 52, L"JavaScript RegEx", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 55, L"JavaScript Start", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 56, L"JavaScript Default", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 57, L"JavaScript Comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 58, L"JavaScript Line Comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 59, L"JavaScript Doc comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 60, L"JavaScript Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 61, L"JavaScript Word", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 62, L"JavaScript Keyword", L"Sample Text",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 63, L"JavaScript Double quoted string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 64, L"JavaScript Single quoted string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 65, L"JavaScript Symbols", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 66, L"JavaScript EOL", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 67, L"JavaScript RegEx", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 70, L"VBS Start", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 71, L"VBS Default", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 72, L"VBS Comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 73, L"VBS Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 74, L"VBS KeyWord", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 75, L"VBS String", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 76, L"VBS Identifier", L"Sample Text",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 77, L"VBS Unterminated string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 80, L"ASP VBS Start", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 81, L"ASP VBS Default", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 82, L"ASP VBS Comment", L"Sample Text",
    RGB(0x00,0x80,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 83, L"ASP VBS Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 84, L"ASP VBS KeyWord", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 85, L"ASP VBS String", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 86, L"ASP VBS Identifier", L"Sample Text",
    RGB(0x00,0x00,0x80), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 87, L"ASP VBS Unterminated string", L"Sample Text",
    RGB(0xC0,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 90, L"Python Start", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 91, L"Python Misc", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 92, L"Python Comment", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 93, L"Python Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 94, L"Python String", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 95, L"Python Single quoted string", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 96, L"Python Keyword", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 97, L"Python Triple quotes", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 98, L"Python Triple double quotes", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 99, L"Python Class definition", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 100, L"Python Function/Method definition", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 101, L"Python Operators", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 102, L"Python Identifiers", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 104, L"PHP complex variable", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 105, L"ASP Python Start", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 106, L"ASP Python Misc", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 107, L"ASP Python Comment", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 108, L"ASP Python Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 109, L"ASP Python String", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 110, L"ASP Python Single quoted string", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 111, L"ASP Python Keyword", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 112, L"ASP Python Triple quotes", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 113, L"ASP Python Triple double quotes", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 114, L"ASP Python Class definition", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 115, L"ASP Python Function/method definition", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 116, L"ASP Python Operators", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 117, L"ASP Python Identifiers", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 118, L"PHP Default", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 119, L"PHP Double quoted String", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 120, L"PHP Single quoted string", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 121, L"PHP Keyword", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 122, L"PHP Number", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 123, L"PHP Variable", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 124, L"PHP Comment", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 125, L"PHP One line comment", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 126, L"PHP variable in double quoted string", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 },

    { 127, L"PHP operator", L"Sample Text",
    RGB(0x00,0x00,0x00), WBSTYLE_DEFAULT_COLOR, 0, 0 }
};
size_t XmlEditorMode::style_cnt_ = countof(styles_);


/* ------------------------------------------------------------------------ */
/*
 *   HTML editor mode.  This is a specialization of the XML editor mode.  
 */
class HtmlEditorMode: public XmlEditorMode
{
public:
    HtmlEditorMode(ITadsWorkbench *itwb) : XmlEditorMode(itwb) { }

    /* get my mode name */
    BSTR STDMETHODCALLTYPE GetModeName()
    {
        return SysAllocString(L"HTML");
    }

    /* get my filter string */
    BSTR STDMETHODCALLTYPE GetFileDialogFilter(BOOL for_save, UINT *priority)
    {
        wchar_t str[] = L"HTML Files (*.htm, *.html)\0*.htm;*.html\0";

        *priority = 1000;
        return SysAllocStringLen(str, sizeof(str)/sizeof(str[0]) - 1);
    }

    /* get my file affinity */
    UINT STDMETHODCALLTYPE GetFileAffinity(const OLECHAR *filename)
    {
        /* the HTML mode recognizes .htm and .html extensions */
        size_t len = wcslen(filename);
        if ((len > 4 && wcsicmp(filename + len - 4, L".htm") == 0)
            || (len > 5 && wcsicmp(filename + len - 5, L".html") == 0))
        {
            /* we matched on extension */
            return 100;
        }
        
        /* not one of ours */
        return 0;
    }

    /* get file type information */
    BOOL STDMETHODCALLTYPE GetFileTypeInfo(UINT index, wb_ftype_info *info)
    {
        switch (index)
        {
        case 0:
            /* XML files */
            info->icon = (HICON)LoadImage(
                G_instance, MAKEINTRESOURCE(ICON_HTML_FILE),
                IMAGE_ICON, LR_DEFAULTSIZE, LR_DEFAULTSIZE, LR_DEFAULTCOLOR);
            info->need_destroy_icon = FALSE;
            wcscpy(info->name, L"HTML File");
            wcscpy(info->desc, L"An HTML Page (HyperText Markup Language)");
            wcscpy(info->defext, L"htm");
            return TRUE;

        default:
            /* no more modes */
            return FALSE;
        }
    }

protected:
    /* create our extension */
    CTadsWorkbenchEditor *create_ext(HWND hwnd);
};


/* ------------------------------------------------------------------------ */
/*
 *   The XML editor customizer 
 */
class XmlEditor: public CTadsWorkbenchEditor
{
public:
    XmlEditor(ITadsWorkbench *itwb, HWND hwnd)
        : CTadsWorkbenchEditor(itwb, hwnd)
    {
    }

    /*
     *   ITadsWorkbenchEditor implementation 
     */

    /* initialize */
    void init()
    {
        /* inherit the default handling */
        CTadsWorkbenchEditor::init();
        
        /* tell Scintilla to use the "html" lexer */
        call_sci(SCI_SETLEXERLANGUAGE, 0, (int)"xml");

        /* the hypertext lexer needs tons of styles */
        call_sci(SCI_SETSTYLEBITS, 7);
        call_sci(SCI_COLOURISE, 0, -1);

        /* add the SGML/DTD keywords */
        call_sci(SCI_SETKEYWORDS, 5, (int)
                 "ELEMENT DOCTYPE ATTLIST ENTITY NOTATION");

        /* send the list of "word" characters */
        call_sci(SCI_SETWORDCHARS, 0, (int)
                 "abcdefghijklmnopqrstuvwxyz"
                 "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                 "0123456789_-");
    }

    /*
     *   Receive notification of an added character 
     */
    void STDMETHODCALLTYPE OnCharAdded(UINT ch)
    {
        /* 
         *   if it's a newline, and auto-indent mode is turned on, do a
         *   simple hanging indent to the same level as the previous line 
         */
        if ((ch == '\n' || ch == '\r') && auto_indent_ != 0)
            hanging_indent();
    }

protected:
};

/* ------------------------------------------------------------------------ */
/*
 *   The HTML editor customizer 
 */
class HtmlEditor: public XmlEditor
{
public:
    HtmlEditor(ITadsWorkbench *itwb, HWND hwnd)
        : XmlEditor(itwb, hwnd)
    {
    }

    /*
     *   ITadsWorkbenchEditor implementation 
     */

    /* initialize */
    void init()
    {
        /* 
         *   inherit the default handling from the base editor class, but
         *   bypass the XML initialization, as we supersede all of that 
         */
        CTadsWorkbenchEditor::init();

        /* tell Scintilla to use the "html" lexer */
        call_sci(SCI_SETLEXERLANGUAGE, 0, (int)"hypertext");

        /* the hypertext lexer needs tons of styles */
        call_sci(SCI_SETSTYLEBITS, 7);
        call_sci(SCI_COLOURISE, 0, -1);

        /* send the lexer the standard list of HTML keywords */
        call_sci(SCI_SETKEYWORDS, 0, (int)
                 /* elements */
                 "a abbr acronym address applet area b base basefont "
                 "bdo big blockquote body br button caption center "
                 "cite code col colgroup dd del dfn dir div dl dt em "
                 "fieldset font form frame frameset h1 h2 h3 h4 h5 h6 "
                 "head hr html i iframe img input ins isindex kbd label "
                 "legend li link map menu meta noframes noscript "
                 "object ol optgroup option p param pre q s samp "
                 "script select small span strike strong style sub sup "
                 "table tbody td textarea tfoot th thead title tr tt u ul "
                 "var xml xmlns "
                 /* attributes */
                 "abbr accept-charset accept accesskey action align alink "
                 "alt archive axis background bgcolor border "
                 "cellpadding cellspacing char charoff charset checked cite "
                 "class classid clear codebase codetype color cols colspan "
                 "compact content coords "
                 "data datafld dataformatas datapagesize datasrc datetime "
                 "declare defer dir disabled enctype event "
                 "face for frame frameborder "
                 "headers height href hreflang hspace http-equiv "
                 "id ismap label lang language leftmargin link longdesc "
                 "marginwidth marginheight maxlength media method multiple "
                 "name nohref noresize noshade nowrap "
                 "object onblur onchange onclick ondblclick onfocus "
                 "onkeydown onkeypress onkeyup onload onmousedown "
                 "onmousemove onmouseover onmouseout onmouseup "
                 "onreset onselect onsubmit onunload "
                 "profile prompt readonly rel rev rows rowspan rules "
                 "scheme scope selected shape size span src standby start style "
                 "summary tabindex target text title topmargin type usemap "
                 "valign value valuetype version vlink vspace width "
                 "text password checkbox radio submit reset "
                 "file hidden image "
                 /* html 5 elements */
                 "article aside calendar canvas card command commandset "
                 "datagrid datatree footer gauge header m menubar menulabel "
                 "nav progress section switch tabbox "
                 /* html 5 attributes */
                 "active command contenteditable ping "
                 /* additional */
                 "public !doctype");

        /* add the javascript keywords */
        call_sci(SCI_SETKEYWORDS, 1, (int)
                 "abstract boolean break byte case catch char class "
                 "const continue debugger default delete do double else "
                 "enum export extends "
                 "final finally float for function goto if implements "
                 "import in instanceof "
                 "int interface long native new package private protected "
                 "public return short static super switch synchronized "
                 "this throw throws "
                 "transient try typeof var void volatile while with");

        /* add the VBScript keywords */
        call_sci(SCI_SETKEYWORDS, 2, (int)
                 "addressof alias and as attribute base begin binary boolean "
                 "byref byte byval call case compare "
                 "const currency date decimal declare defbool defbyte defint "
                 "deflng defcur defsng defdbl defdec "
                 "defdate defstr defobj defvar dim do double each else "
                 "elseif empty end enum eqv erase error "
                 "event exit explicit false for friend function get gosub "
                 "goto if imp implements in input integer "
                 "is len let lib like load lock long loop lset me mid midb "
                 "mod new next not nothing null object "
                 "on option optional or paramarray preserve print private "
                 "property public raiseevent randomize "
                 "redim rem resume return rset seek select set single "
                 "static step stop string sub then time to "
                 "true type typeof unload until variant wend while with "
                 "withevents xor");

        /* add the Python keywords */
        call_sci(SCI_SETKEYWORDS, 3, (int)
                 "and assert break class continue def del elif "
                 "else except exec finally for from global if import "
                 "in is lambda None not or pass print raise return try "
                 "while yield");

        /* add the PHP keywords */
        call_sci(SCI_SETKEYWORDS, 4, (int)
                 "and array as bool boolean break case cfunction class "
                 "const continue declare "
                 "default die directory do double echo else elseif empty "
                 "enddeclare endfor "
                 "endforeach endif endswitch endwhile eval exit extends "
                 "false float for "
                 "foreach function global if include include_once int "
                 "integer isset list new "
                 "null object old_function or parent print real require "
                 "require_once resource "
                 "return static stdclass string switch true unset use "
                 "var while xor __class__ "
                 "__file__ __function__ __line__ __sleep __wakeup");

        /* add the SGML/DTD keywords */
        call_sci(SCI_SETKEYWORDS, 5, (int)
                 "ELEMENT DOCTYPE ATTLIST ENTITY NOTATION");

        /* send the list of "word" characters */
        call_sci(SCI_SETWORDCHARS, 0, (int)
                 "abcdefghijklmnopqrstuvwxyz"
                 "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                 "0123456789_-");
    }
};




/* ------------------------------------------------------------------------ */
/*
 *   The extension object.
 */
class HtmlAddIn: public CTadsWorkbenchExtension
{
public:
    HtmlAddIn(ITadsWorkbench *itwb)
        : CTadsWorkbenchExtension(itwb)
    {
    }

    /*
     *   ITadsWorkbenchExtension implementation 
     */

    /* get my nth editor mode */
    ITadsWorkbenchEditorMode * STDMETHODCALLTYPE GetMode(UINT index)
    {
        switch (index)
        {
        case 0:
            return new HtmlEditorMode(itwb_);

        case 1:
            return new XmlEditorMode(itwb_);

        default:
            /* no other modes */
            return 0;
        }
    }

protected:
};


/* ------------------------------------------------------------------------ */
/*
 *   Additional Mode object implementation 
 */

/* create the HTML extension */
CTadsWorkbenchEditor *HtmlEditorMode::create_ext(HWND hwnd)
{
    return new HtmlEditor(itwb_, hwnd);
}

/* create the XML extension */
CTadsWorkbenchEditor *XmlEditorMode::create_ext(HWND hwnd)
{
    return new XmlEditor(itwb_, hwnd);
}


/* ------------------------------------------------------------------------ */
/*
 *   Connect.  Workbench calls this at start-up to initialize the DLL and
 *   exchange interface pointers.  
 */
__declspec(dllexport) IUnknown *twb_connect(ITadsWorkbench *itwb)
{
    /* create and return our extension object */
    return new HtmlAddIn(itwb);
}
