/* Copyright (c) 2006 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  wb_addin.h - TADS Workbench Add-In helpers
Function
  This defines some helper classes that make it easier to write a
  TADS Workbench extension add-in.
Notes
  
Modified
  10/18/06 MJRoberts  - Creation
*/

#ifndef WB_ADDIN_H
#define WB_ADDIN_H

#include "itadsworkbench.h"
#include <scintilla.h>


/* ------------------------------------------------------------------------ */
/*
 *   Auto-indent modes 
 */
#define TWB_AUTOINDENT_NONE   0                        /* no auto-indenting */
#define TWB_AUTOINDENT_SIMPLE 1     /* simple: use same indent as prior line */
#define TWB_AUTOINDENT_SYNTAX 2              /* full syntax-based indenting */

/*
 *   Comment auto-formatting modes 
 */
#define TWB_FMTCOMMENT_OFF    0              /* do not auto-format comments */
#define TWB_FMTCOMMENT_ON     1            /* automatically format comments */


/* ------------------------------------------------------------------------ */
/*
 *   Utility routines 
 */

/* number of elements in an array */
#define countof(x) (sizeof(x)/sizeof(x[0]))

/* "safe" wide strcpy - don't overflow, always null-terminate */
static void safe_wcscpy(wchar_t *dst, size_t dstcnt, const wchar_t *src)
{
    size_t copycnt;

    /* if the destination buffer is zero-size, we can't do anything */
    if (dstcnt == 0)
        return;

    /* get the source length */
    copycnt = wcslen(src);

    /* limit to the available destination space less null termination */
    if (copycnt > dstcnt - 1)
        copycnt = dstcnt - 1;

    /* copy the characters */
    memcpy(dst, src, copycnt * sizeof(*dst));

    /* null-terminate */
    dst[copycnt] = L'\0';
}

/* ------------------------------------------------------------------------ */
/* 
 *   Private style description structure.  We use this to initialize a static
 *   array of style data, for ITadsWorkbenchEditorMode::GetStyleInfo()
 *   implementations.  
 */
struct style_info
{
    int sci_index;
    const OLECHAR *title;
    const OLECHAR *sample;
    COLORREF fgcolor;
    COLORREF bgcolor;
    UINT styles;
    int keyword_set;
};


/* ------------------------------------------------------------------------ */
/*
 *   Scintilla fast-path invoker 
 */
class SciInvoker
{
public:
    SciInvoker(HWND hwnd) { init(hwnd); }
    SciInvoker(SciInvoker &sci) { init(sci); }

    void init(HWND hwnd)
    {
        /* remember the window handle */
        hwnd_ = hwnd;

        /* set up the direct-call interface pointers */
        call_sci_ = (int (__cdecl *)(void *, int, int, int))SendMessage(
            hwnd, SCI_GETDIRECTFUNCTION, 0, 0);
        call_sci_ctx_ = (void *)SendMessage(
            hwnd, SCI_GETDIRECTPOINTER, 0, 0);

        /* look up the custom message names */
        msg_create_spot = RegisterWindowMessage("w32tdb.sci.CreateSpot");
        msg_delete_spot = RegisterWindowMessage("w32tdb.sci.DeleteSpot");
        msg_get_spot_pos = RegisterWindowMessage("w32tdb.sci.GetSpotPos");
        msg_set_spot_pos = RegisterWindowMessage("w32tdb.sci.SetSpotPos");
    }

    void init(SciInvoker &sci)
    {
        call_sci_ = sci.call_sci_;
        call_sci_ctx_ = sci.call_sci_ctx_;
    }

    /* invoke Scintilla through the direct-call interface */
    int invoke(int msg, int a = 0, int b = 0)
        { return call_sci_(call_sci_ctx_, msg, a, b); }

    /* call the various Workbench custom message handlers */
    LPARAM create_spot(int pos)
        { return SendMessage(hwnd_, msg_create_spot, pos, 0); }
    void delete_spot(LPARAM spot)
        { SendMessage(hwnd_, msg_delete_spot, 0, spot); }
    int get_spot_pos(LPARAM spot)
        { return SendMessage(hwnd_, msg_get_spot_pos, 0, spot); }
    void set_spot_pos(LPARAM spot, int pos)
        { SendMessage(hwnd_, msg_set_spot_pos, pos, spot); }

private:
    /* the Scintilla window handle */
    HWND hwnd_;
        
    /* Scintilla direct function interface */
    int (__cdecl *call_sci_)(void *, int, int, int);
    void *call_sci_ctx_;

    /* custom Workbench extension messages in the Scintilla window */
    UINT msg_create_spot;
    UINT msg_delete_spot;
    UINT msg_get_spot_pos;
    UINT msg_set_spot_pos;
};


/* ------------------------------------------------------------------------ */
/*
 *   Scintilla line buffer.  This encapsulates a line of text from Scintilla.
 */
class SciLine
{
public:
    SciLine(SciInvoker &sci)
        : sci_(sci)
    {
        /* initialize */
        init();
    }
    SciLine(SciInvoker &sci, int l)
        : sci_(sci)
    {
        /* initialize */
        init();

        /* retrieve the initial line */
        retrieve_line(l);
    }

    ~SciLine()
    {
        /* delete our buffer, if we have one */
        if (txt != 0)
            delete [] txt;
    }

    /* get the column of the given position in our buffer */
    int getcol(char *p)
        { return sci_.invoke(SCI_GETCOLUMN, pos + (p - txt)); }

    /* get the text at the given column */
    char *getcoltxt(int col)
        { return txt + (sci_.invoke(SCI_FINDCOLUMN, lin, col) - pos); }

    /* get the text for a given buffer position */
    char *getpostxt(int p)
    {
        if (p < pos)
            return txt;
        else if (p > pos + len)
            return txt + len;
        else
            return txt + (p - pos);
    }

    /* get the buffer position for a given text pointer */
    int gettxtpos(const char *p)
    {
        if (p >= txt && p < txt + len)
            return pos + (p - txt);
        else if (p >= txt + len)
            return pos + len;
        else
            return pos;
    }

    /* 
     *   Retrieve the given line number.  This fills in 'txt' with the text
     *   of the line, stripped of newlines.  The text is null-terminated.  
     */
    void retrieve_line(int l)
    {
        char *p;

        /* note the new line number */
        lin = l;

        /* get the length */
        len = sci_.invoke(SCI_LINELENGTH, l);

        /* delete any existing buffer */
        if (txt != 0)
            delete [] txt;

        /* allocate space */
        txt = new char[len + 1];

        /* note the position of the line */
        pos = sci_.invoke(SCI_POSITIONFROMLINE, l);

        /* retrieve it */
        sci_.invoke(SCI_GETLINE, l, (int)txt);

        /* null-terminate it */
        txt[len] = '\0';

        /* remove trailing nulls and newlines */
        for (p = txt + len ;
             p > txt && (*(p-1) == '\0' || *(p-1) == '\n' || *(p-1) == '\r') ;
             *--p = '\0', --len) ;
    }

    /* our text buffer */
    char *txt;

    /* the length of the text (not counting the null terminator) */
    int len;

    /* the current line number */
    int lin;

    /* the text position of the line */
    int pos;

protected:
    void init()
    {
        /* we don't have a line of text yet */
        lin = -1;
        pos = -1;
        txt = 0;
    }

    /* our Scintilla invoker */
    SciInvoker &sci_;
};


/* ------------------------------------------------------------------------ */
/*
 *   Common base class for ITadsWorkbenchEditorMode implementations. 
 */
class CTadsWorkbenchEditorMode: public ITadsWorkbenchEditorMode
{
public:
    CTadsWorkbenchEditorMode(ITadsWorkbench *itwb)
    {
        /* remember my ITadsWorkbench object */
        itwb->AddRef();
        itwb_ = itwb;

        /* set an initial reference count of 1 on behalf of our caller */
        refcnt_ = 1;
    }

    virtual ~CTadsWorkbenchEditorMode()
    {
        /* done with our ITadsWorkbench object */
        itwb_->Release();
    }

    /* IUnknown reference-counting */
    ULONG STDMETHODCALLTYPE AddRef() { return ++refcnt_; }
    ULONG STDMETHODCALLTYPE Release()
    {
        ULONG ret;
        if ((ret = --refcnt_) == 0)
            delete this;
        return ret;
    }

    /* IUnknown interface query */
    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ifc)
    {
        /* check for our known interfaces */
        if (iid == IID_IUnknown)
            *ifc = (void *)(IUnknown *)this;
        else if (iid == IID_ITadsWorkbenchEditorMode)
            *ifc = (void *)(ITadsWorkbenchEditorMode *)this;
        else
            return E_NOINTERFACE;

        /* add a reference on behalf of the caller, per COM conventions */
        AddRef();

        /* success */
        return S_OK;
    }

    /*
     *   ITadsWorkbenchEditorMode methods 
     */

    /* get my default style settings */
    void STDMETHODCALLTYPE GetStyleInfo(UINT idx, wb_style_info *info)
    {
        /* if the index is out of range, just say so */
        if (idx >= get_style_count())
            return;

        /* get this style definer */
        const style_info *s = &get_style_array()[idx];

        /* if it's undefined, fail */
        if (s == 0)
            return;

        /* 
         *   Fill in the information.  We use the default font for all of our
         *   styles, so we can just leave the name empty and set the point
         *   size to zero.  
         */
        info->sci_index = s->sci_index;
        info->font_name[0] = '\0';
        info->ptsize = 0;
        info->fgcolor = s->fgcolor;
        info->bgcolor = s->bgcolor;
        info->styles = s->styles;
        info->keyword_set = s->keyword_set;
        safe_wcscpy(info->title, countof(info->title), s->title);
        safe_wcscpy(info->sample, countof(info->sample), s->sample);
    }

    /* get the number of styles we use */
    UINT STDMETHODCALLTYPE GetStyleCount()
    {
        /* return the number of array elements */
        return get_style_count();
    }

    /* create an instance of our extension */
    ITadsWorkbenchEditor * STDMETHODCALLTYPE CreateEditorExtension(HWND hwnd);

protected:
    /* 
     *   Each subclass must override this to return a new object of the
     *   appropriate CTadsWorkbenchEditor subclass.  The implementation is
     *   usually just something like 'return new MySubclass(itwb_, hwnd)'.  
     */
    virtual class CTadsWorkbenchEditor *create_ext(HWND hwnd) = 0;

    /*
     *   Get an entry from my array of style_info elements, and get the
     *   number of elements in the array.  Subclasses can use our default
     *   implementation of GetStyleInfo() simply by defining a static array
     *   of style_info elements, and returning that static array address
     *   here.  Alternatively, the subclass can provide its own completely
     *   custom implementation, without defining a style_info array at all,
     *   by overriding GetStyleInfo() and GetStyleCount(); the subclass can
     *   then provide dummy implementations of these methods (i.e., just
     *   return 0 from both of these).  
     */
    virtual const style_info *get_style_array() = 0;
    virtual UINT get_style_count() = 0;

    /* reference counter */
    ULONG refcnt_;

    /* my Workbench interface */
    ITadsWorkbench *itwb_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Common base class for ITadsWorkbenchEditor implementations. 
 */
class CTadsWorkbenchEditor: public ITadsWorkbenchEditor
{
public:
    CTadsWorkbenchEditor(ITadsWorkbench *itwb, HWND hwnd)
        : sci_(hwnd)
    {
        /* count an initial reference on behalf of the caller */
        refcnt_ = 1;

        /* remember the Workbench interface */
        itwb->AddRef();
        itwb_ = itwb;

        /* remember our Scintilla window */
        hwnd_ = hwnd;
    }

    virtual ~CTadsWorkbenchEditor()
    {
        /* release our objects */
        itwb_->Release();
    }

    /*
     *   Initialize.  The editor-mode object's CreateEditorExtension must
     *   call this after creating the object.  
     */
    virtual void init()
    {
        /* note our initial configuration settings */
        OnSettingsChange();
    }

    /* IUnknown reference-counting */
    ULONG STDMETHODCALLTYPE AddRef() { return ++refcnt_; }
    ULONG STDMETHODCALLTYPE Release()
    {
        ULONG ret;
        if ((ret = --refcnt_) == 0)
            delete this;
        return ret;
    }

    /* IUnknown interface query */
    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ifc)
    {
        /* check for our known interfaces */
        if (iid == IID_IUnknown)
            *ifc = (void *)(IUnknown *)this;
        else if (iid == IID_ITadsWorkbenchEditor)
            *ifc = (void *)(ITadsWorkbenchEditor *)this;
        else
            return E_NOINTERFACE;

        /* add a reference on behalf of the caller, per COM conventions */
        AddRef();

        /* success */
        return S_OK;
    }

    /* 
     *   ITadsWorkbenchEditor methods
     */

    /* 
     *   Receive notification of a Workbench configuration settings change.
     *   Subclasses should override if they need to refresh any property
     *   settings beyond the styles and the other properties we refresh here.
     *   In particular, editor modes that use custom keyword lists must
     *   refresh those lists in an override.  
     */
    void STDMETHODCALLTYPE OnSettingsChange()
    {
        /*   
         *   Get the auto-indent and format-comments modes from Workbench.
         *   The default auto-indent mode is "full syntax indenting."  The
         *   default comment formatting mode is "on."  These are global
         *   options settings.  
         */
        auto_indent_ = itwb_->GetPropValInt(
            TRUE, L"src win options", L"autoindent",
            TWB_AUTOINDENT_SYNTAX);
        format_comments_ = itwb_->GetPropValInt(
            TRUE, L"src win options", L"autoformat comments",
            TWB_FMTCOMMENT_ON);
    }

    /*
     *   Message filter.  Our default implementation simply passes the
     *   message directly to the default handler, and returns the result.  
     */
    LRESULT STDMETHODCALLTYPE MsgFilter(
        HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam,
        ITadsWorkbenchMsgHandler *def_handler)
    {
        /* pass the message directly to the default handler */
        return def_handler->CallMsgHandler(hwnd, msg, wparam, lparam);
    }

    /* 
     *   Command status updater.  By default, we don't add any special
     *   commands of our own or override any Workbench commands, so we simply
     *   return FALSE here to tell Workbench to carry on with the default
     *   status check.  Subclasses should override this to indicate the
     *   status for extended commands they define.  
     */
    BOOL STDMETHODCALLTYPE GetCommandStatus(
        UINT command_id, UINT *status, HMENU menu, UINT menu_index)
    {
        return FALSE;
    }

    /* 
     *   Carry out a command.  By default, we don't add any commands of our
     *   own or override any Workbench commands, so we simply return FALSE to
     *   let Workbench carry out the default processing. 
     */
    BOOL STDMETHODCALLTYPE DoCommand(UINT command_id)
    {
        return FALSE;
    }

protected:
    /* 
     *   Call Scintilla.  This uses the fast-path interface to the Scintilla
     *   window, which the Scintilla documentation describes as much faster
     *   than the SendMessage() interface.  It is otherwise similar: just
     *   pass in a SCI_xxx message ID for 'msg', along with any parameters
     *   the message requires.  
     */
    int call_sci(int msg, int a = 0, int b = 0)
        { return sci_.invoke(msg, a, b); }

    /* call the various Workbench custom message handlers */
    LPARAM create_spot(int pos) { return sci_.create_spot(pos); }
    void delete_spot(LPARAM spot) { return sci_.delete_spot(spot); }
    int get_spot_pos(LPARAM spot) { return sci_.get_spot_pos(spot); }
    void set_spot_pos(LPARAM spot, int pos) { sci_.set_spot_pos(spot, pos); }

    /* 
     *   get the standard newline sequence for the current file - this
     *   returns \r\n, \n, or \r, according to the newline mode that the
     *   underlying Scintilla document uses 
     */
    const char *get_nl()
    {
        static const char *nl[] = { "\r\n", "\r", "\n" };

        /* ask Scintilla for its current newline mode */
        int mode = call_sci(SCI_GETEOLMODE);

        /* if we recognize the mode, return its newline sequence */
        if (mode >= 0 && mode < countof(nl))
            return nl[mode];
        else
            return "\n";
    }

    /* 
     *   Do simple hanging indentation - indent the current line to the same
     *   level as the previous line, and leave the cursor at the indentation
     *   point in the current line.  
     */
    void hanging_indent()
    {
        /* get the current line */
        int pos = call_sci(SCI_GETCURRENTPOS);
        int lin = call_sci(SCI_LINEFROMPOSITION, pos);

        /* 
         *   if we're not at the first line, indent to the previous line's
         *   indentation position 
         */
        if (lin > 0)
        {
            /* set the same indentation as the previous line */
            int ind = call_sci(SCI_GETLINEINDENTATION, lin - 1);
            call_sci(SCI_SETLINEINDENTATION, lin, ind);

            /* go to the indent position */
            pos = call_sci(SCI_GETLINEINDENTPOSITION, lin);
            call_sci(SCI_GOTOPOS, pos);
        }
    }

    /* the Scintilla window handle */
    HWND hwnd_;

    /* the Scintilla invoker */
    SciInvoker sci_;

    /* my callback interface to Workbench */
    ITadsWorkbench *itwb_;

    /* COM reference count */
    ULONG refcnt_;


    /*
     *   Cached configuration settings.  We retrieve this at initialization,
     *   and refresh them whenever we're notified of a settings change.  
     */

    /* auto-indent mode (TWB_AUTOINDENT_xxx) */
    int auto_indent_;

    /* auto-format comments mode (TWB_FMTCOMMENT_xxx) */
    int format_comments_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Additional editor mode implementation 
 */

ITadsWorkbenchEditor * STDMETHODCALLTYPE CTadsWorkbenchEditorMode::
   CreateEditorExtension(HWND hwnd)
{
    CTadsWorkbenchEditor *ext;

    /* create the extension object */
    ext = create_ext(hwnd);
    
    /* initialize it */
    ext->init();
    
    /* return it */
    return ext;
}


/* ------------------------------------------------------------------------ */
/*
 *   Base class for extension objects 
 */
class CTadsWorkbenchExtension: public ITadsWorkbenchExtension
{
public:
    CTadsWorkbenchExtension(ITadsWorkbench *itwb)
    {
        /* remember the Workbench interface */
        itwb->AddRef();
        itwb_ = itwb;

        /* count an initial reference on behalf of the caller */
        refcnt_ = 1;
    }

    virtual ~CTadsWorkbenchExtension()
    {
        /* release our Workbench interface */
        itwb_->Release();
    }

    /* IUnknown reference-counting */
    ULONG STDMETHODCALLTYPE AddRef() { return ++refcnt_; }
    ULONG STDMETHODCALLTYPE Release()
    {
        ULONG ret;
        if ((ret = --refcnt_) == 0)
            delete this;
        return ret;
    }

    /* IUnknown interface query */
    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ifc)
    {
        /* check for our known interfaces */
        if (iid == IID_IUnknown)
            *ifc = (void *)(IUnknown *)this;
        else if (iid == IID_ITadsWorkbenchExtension)
            *ifc = (void *)(ITadsWorkbenchExtension *)this;
        else
            return E_NOINTERFACE;

        /* add a reference on behalf of the caller, per COM conventions */
        AddRef();

        /* success */
        return S_OK;
    }

protected:
    /* my callback interface to Workbench */
    ITadsWorkbench *itwb_;

    /* my reference count */
    ULONG refcnt_;
};

#endif /* WB_ADDIN_H */
